
#ifndef PLS_CONT_MANAGER_H_
#define PLS_CONT_MANAGER_H_

#include <memory>
#include <utility>
#include <array>

#include "pls/internal/data_structures/aligned_stack.h"
#include "pls/internal/scheduling/continuation.h"

namespace pls {
namespace internal {
namespace scheduling {

class cont_manager {
 public:
  template<size_t NUM_CONTS, size_t MAX_CONT_SIZE>
  explicit cont_manager(data_structures::aligned_stack &cont_storage) {
    // First node is currently active and our local start
    start_node_ = active_node_ = init_cont_node<MAX_CONT_SIZE>(cont_storage, nullptr, nullptr);

    // Build up chain after it
    continuation_node *current_node = start_node_;
    for (size_t i = 1; i < NUM_CONTS; i++) {
      continuation_node *next_node = init_cont_node<MAX_CONT_SIZE>(cont_storage, start_node_, current_node);
      current_node->set_prev(next_node);
    }
  };

  continuation_node *fast_path_get_next() {
    active_node_ = active_node_->get_next();
    return active_node_;
  }
  void fast_path_return() {
    active_node_ = active_node_->get_prev();
  }

 private:
  template<size_t MAX_CONT_SIZE>
  static continuation_node *init_cont_node(data_structures::aligned_stack &cont_storage,
                                           continuation_node *cont_chain_start,
                                           continuation_node *prev) {
    // Represents one cont node and its corresponding memory buffer (as one continuous block of memory).
    using cont_node_memory_pair = std::pair<continuation_node,
                                             std::array<char, MAX_CONT_SIZE - sizeof(continuation_node)>>;
    char *pair_memory = cont_storage.push_bytes<cont_node_memory_pair>();
    char *cont_node_address = pair_memory;
    char *cont_node_memory_address = pair_memory + sizeof(continuation_node);

    return new(cont_node_address) continuation_node(cont_node_memory_address, cont_chain_start, prev);
  }

 private:
  continuation_node *start_node_;
  continuation_node *active_node_;
};

template<size_t NUM_CONTS, size_t MAX_CONT_SIZE>
class static_cont_manager {
 public:
  static_cont_manager() : static_cont_storage_{}, cont_manager_{NUM_CONTS, MAX_CONT_SIZE, static_cont_storage_} {}
  cont_manager &get_cont_manager() { return cont_manager_; }

 private:
  data_structures::static_aligned_stack<NUM_CONTS * MAX_CONT_SIZE> static_cont_storage_;
  cont_manager cont_manager_;
};

}
}
}
#endif //PLS_CONT_MANAGER_H_
