
#ifndef PLS_TASK_H
#define PLS_TASK_H

#include "pls/internal/scheduling/thread_state.h"
#include "pls/internal/scheduling/continuation.h"

namespace pls {
namespace internal {
namespace scheduling {

/**
 * A task to be executed by the runtime system.
 * Tasks are guaranteed to be executed exactly once.
 *
 * Override the execute_internal() method for your custom code.
 */
class base_task {
  friend class scheduler;

 protected:
  base_task() = default;

  /**
   * Overwrite this with the actual behaviour of concrete tasks.
   */
  virtual void execute_internal() = 0;

 private:
  void execute() {
    // TODO: Figure out slow path execution
    execute_internal();
  }
};

template<typename F, typename R1, typename R2, typename CF>
class task : public base_task {
 public:
  template<typename FARG>
  explicit task(FARG &&function, continuation<R1, R2, CF> *continuation)
      : base_task{}, function_{std::forward<FARG>(function)}, continuation_{continuation} {}

  void execute_internal() override {
    continuation_->store_result_2(function_());
    // TODO: Properly notify continuation on slow path
  }

 private:
  F function_;
  continuation<R1, R2, CF> *continuation_;
};

}
}
}

#endif //PLS_TASK_H
